/*Copyright (C) 2009  Annette Bieniusa.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package avl;

import java.util.HashMap;
import java.util.Map;

import dstm.*;

/**
 * Implementation of a generic AVLTree library.
 * @author Annette Bieniusa
 * @author Arie Middelkoop
 */

class AVLTree<K extends Comparable<K>,V> extends LOC<AVLTree<K,V>> {
	private K 					_key;
	private V 					_value;
	private int 				_height;
	private GAO<AVLTree<K,V>> 	_left;
	private GAO<AVLTree<K,V>> 	_right;
	
	public AVLTree() {
		_key	= null;
		_value	= null;
		_height = 0;
		_left	= null;
		_right	= null;
	}
	
	public AVLTree(K key, V value) {
		assert(key != null);

		_key 	= key;
		_value 	= value;
		_height = 1;
		_left	= null;
		_right  = null;
	}
	
	public K getKey() {
		assert(_key != null);
		return _key;
	}
	
	public V getValue() {
		return _value;
	}
	
	
	// ----------------------------------------------------------
	//  Search
	// ----------------------------------------------------------

	public V find(K key, Transaction trans) {
		assert(key != null);
		assert(trans != null);
		
		if (_key == null) {
			return null;
		}
		assert(_key != null);

		int comp = key.compareTo(_key);
		
		if (comp > 0 && _right != null) {
			AVLTree<K,V> right = trans.read(_right);
			return right.find(key, trans);
		} else if (comp < 0 && _left != null) {
			AVLTree<K,V> left = trans.read(_left);
			return left.find(key, trans);
		} else if (comp == 0 && _height > 0) {
			return _value;
		} else {
			return null;
		}
	}
	
	public boolean contains(K key, Transaction trans) {
		return find(key, trans) != null;
	}
	
	
	// ----------------------------------------------------------
	//   Insertion
	// ----------------------------------------------------------
	
	public void insert(K key, V value, Transaction trans) {
		assert(key != null);
		assert(trans != null);

		if (_height == 0) {
			_key 	= key;
			_value 	= value;
			_height = 1;
			trans.write(this);
		} else {
			int comp = key.compareTo(_key);
			
			if (comp == 0) {
				_value = value;
				trans.write(this);
			} else if (comp > 0) {
				if (_right != null) {
					AVLTree<K,V> right = trans.read(_right);
					right.insert(key, value, trans);
				} else {
					AVLTree<K,V> right = new AVLTree<K,V>(key, value);
					_right = GAO.create(right,trans);
					trans.write(this);
				}
			} else if (comp < 0) {
				if (_left != null) {
					AVLTree<K,V> left = trans.read(_left);
					left.insert(key, value, trans);
				} else {
					AVLTree<K,V> left = new AVLTree<K,V>(key, value);
					_left = GAO.create(left,trans);
					trans.write(this);
				}
			}
		}
		
		updateHeight(trans);
		rebalance(trans);
	}
	
	private void updateHeight(Transaction trans) {
		if (_height != 0) {
			int height = 1 + Math.max(leftHeight(trans), rightHeight(trans));
			if (height != _height) {
				_height = height;
				trans.write(this);
			}
		}
	}
	
	private int leftHeight(Transaction trans) {
		if (_left != null) {
			return trans.read(_left)._height;
		} else {
			return 0;
		}
	}
	
	private int rightHeight(Transaction trans) {
		if (_right != null) {
			return trans.read(_right)._height;
		} else {
			return 0;
		}
	}
	
	
	// ----------------------------------------------------------
	//   Deletion
	// ----------------------------------------------------------
	
	public void delete(K key, Transaction trans) {
		assert(key != null);
		assert(trans != null);
		
		if (_height != 0) {
			int comp = key.compareTo(_key);
			
			if (comp > 0 && _right != null) {
				trans.read(_right).delete(key, trans);
			} else if (comp < 0 && _left != null) {
				trans.read(_left).delete(key, trans);
			} else if (comp == 0) {
				if (_left == null && _right == null) {
					_height = 0;
					_key    = null;
					_value  = null;
					trans.write(this);
				} else if (_left != null) {
					AVLTree<K,V> left = trans.read(_left);
					AVLTree<K,V> largest = left.largest(trans);
					_key = largest._key;
					_value = largest._value;
					trans.write(this);
					left.delete(largest._key, trans);
				} else {
					assert(_right != null);
					AVLTree<K,V> right = trans.read(_right);
					AVLTree<K,V> smallest = right.smallest(trans);
					_key = smallest._key;
					_value = smallest._value;
					trans.write(this);
					right.delete(smallest._key, trans);
				}
			}
			
			if (_right != null) {
				AVLTree<K,V> right = trans.read(_right);
				if (right._height == 0) {
					_right = null;
					trans.write(this);
				}
			}
			
			if (_left != null) {
				AVLTree<K,V> left = trans.read(_left);
				if (left._height == 0) {
					_left = null;
					trans.write(this);
				}
			}
			
			updateHeight(trans);
			rebalance(trans);
		}
	}
	
	private AVLTree<K,V> smallest(Transaction trans) {
        if (_left == null)
            return this;
        else
            return trans.read(_left).smallest(trans);
    }
	
	private AVLTree<K,V> largest(Transaction trans) {
        if (_right == null)
            return this;
        else
            return trans.read(_right).largest(trans);
    }

	public Map<K,V> getMap(Transaction trans) {
		HashMap<K,V> result = new HashMap<K,V>();
		
		if (_height > 0)
			result.put(_key, _value);
		
		if (_left != null) {
			result.putAll(trans.read(_left).getMap(trans));
		}
		
		if (_right != null) {
			result.putAll(trans.read(_right).getMap(trans));
		}
		
		return result;
	}
	
	// ----------------------------------------------------------
	//   Rebalancing
	// ----------------------------------------------------------
	
	private void rebalance(Transaction trans) {
		if (leftHeight(trans) > rightHeight(trans) + 1) {
			assert(_left != null);
			//assert(_right != null);
			
			AVLTree<K,V> left = trans.read(_left);
			if (left.leftHeight(trans) > left.rightHeight(trans))
				rebalanceLeftLeft(trans);
			else
				rebalanceLeftRight(trans);

			updateTwoLevelHeight(trans);
		} else if (rightHeight(trans) > leftHeight(trans) + 1) {
			//assert(_left != null);
			assert(_right != null);
			
			AVLTree<K,V> right = trans.read(_right);
			if (right.leftHeight(trans) > right.rightHeight(trans))
				rebalanceRightLeft(trans);
			else
				rebalanceRightRight(trans);

			updateTwoLevelHeight(trans);
		}
		
		assert(isBalanced(trans));
	}
	
	private void updateTwoLevelHeight(Transaction trans)
    {
		assert(_left != null);
		assert(_right != null);
		
        trans.read(_left).updateHeight(trans);
        trans.read(_right).updateHeight(trans);
        updateHeight(trans);
    }

	private boolean isBalanced(Transaction trans) {
		return Math.abs(leftHeight(trans) - rightHeight(trans)) <= 1;
	}
	

	// ----------------------------------------------------------
	//   Rotations
	// ----------------------------------------------------------
	
	private void swapValues(AVLTree<K,V> other, Transaction trans) {
		K origKey   = _key;
		V origValue = _value;
		
		_key   = other._key;
		_value = other._value;
		
		other._key   = origKey;
		other._value = origValue;
		
		trans.write(this);
		trans.write(other);
	}

	private void rebalanceLeftLeft(Transaction trans)
    {
        AVLTree<K,V> p, q, r;
        GAO<AVLTree<K,V>> a, b, c, d, gp, gr, gq;

        // pattern match (p, q, r != null)
        gr = this.getGao();
        r = this;
        d = r._right;
        gq = r._left;
        q = trans.read(gq);
        c = q._right;
        gp = q._left;
        p = trans.read(gp);
        b = p._right;
        a = p._left;

        // make q the new root
        swapValues(q, trans);
        r  = q;
        gr = gq;
        q  = this;
        gq = this.getGao();

        // reconfigure tree
        q._left  = gp;
        q._right = gr;
        p._left  = a;
        p._right = b;
        r._left  = c;
        r._right = d;
        
        trans.write(p);
        trans.write(q);
        trans.write(r);
    }
	
	private void rebalanceLeftRight(Transaction trans)
    {
        AVLTree<K,V> p, q, r;
        GAO<AVLTree<K,V>> a, b, c, d, gq, gr, gp;

        // pattern match (p, q, r != null)
        r = this;
        gr = this.getGao();
        d = r._right;
        gq = r._left;
        q = trans.read(gq);
        c = q._left;
        gp = q._right;
        p = trans.read(gp);
        b = p._right;
        a = p._left;

        // make p the new root
        swapValues(p, trans);
        r  = p;
        gr = gp;
        p  = this;
        gp = this.getGao();

        // reconfigure tree
        p._left  = gq;
        p._right = gr;
        q._left  = c;
        q._right = a;
        r._left  = b;
        r._right = d;
        
        trans.write(p);
        trans.write(q);
        trans.write(r);
    }
	
	private void rebalanceRightLeft(Transaction trans)
    {
		AVLTree<K,V> p, q, r;
		GAO<AVLTree<K,V>> a, b, c, d, gr, gp, gq;

        // pattern match (p, q, r != null)
        r = this;
        gr = this.getGao();
        d = r._left;
        gq = r._right;
        q = trans.read(gq);
        c = q._right;
        gp = q._left;
        p = trans.read(gp);
        b = p._right;
        a = p._left;

        // make p the new root
        swapValues(p, trans);
        r  = p;
        gr = gp;
        p  = this;
        gp = this.getGao();

        // reconfigure tree
        p._left  = gr;
        p._right = gq;
        r._left  = d;
        r._right = a;
        q._left  = b;
        q._right = c;
        
        trans.write(p);
        trans.write(q);
        trans.write(r);
    }
	
	private void rebalanceRightRight(Transaction trans)
    {
        AVLTree<K,V> p, q, r;
        GAO<AVLTree<K,V>> a, b, c, d, gp, gq, gr;

        // pattern match (p, q, r != null)
        r = this;
        gr = this.getGao();
        d = r._left;
        gq = r._right;
        q = trans.read(gq);
        c = q._left;
        gp = q._right;
        p = trans.read(gp);
        b = p._right;
        a = p._left;

        // make q the new root
        swapValues(q, trans);
        r  = q;
        gr = gq;
        q  = this;
        gq = this.getGao();

        // reconfigure tree
        q._left  = gr;
        q._right = gp;
        r._left  = d;
        r._right = c;
        p._left  = a;
        p._right = b;
        
        trans.write(p);
        trans.write(q);
        trans.write(r);
    }
}
