/*Copyright (C) 2009  Annette Bieniusa.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package avl;

import java.io.FileWriter;
import java.io.IOException;
import dstm.*;
import dstm.Runtime;

public class Benchmark {
	static int NUM_INIT = 100;
	static int NUM_THREADS = 0;
	static int RANGE = 1024;//Integer.MAX_VALUE;
	static int NUM_OPS = 0;
	static int NUM_OPS_PER_THREAD = 0;
	static double LOOKUPS = 1.0;
	static double UPDATES = 0.0;

	private static class SortedAdd extends Thread {
		private GAO<AVLTree<Integer, Integer>> tree;

		double read;
		double insert;
		double delete;
		BenchmarkData bd;

		SortedAdd(GAO<AVLTree<Integer, Integer>> tree, double read, double insert,
				double delete, BenchmarkData bd) {
			this.tree = tree;
			this.read = read;
			this.insert = insert;
			this.delete = delete;
			this.bd = bd;
		}

		public void run() {
			for (int i = 0; i < NUM_OPS_PER_THREAD; i++) {
					double rand = Math.random();
					int key = (int) (Math.random() * RANGE);
					boolean done = false;
					while(!done) {
						try {
							Transaction trans; 
							if (rand < read) {
								// System.out.println(Thread.currentThread().getName()+":
								// Search for "+key);
									trans = Runtime.createTransacation("find "+key);
									trans.read(tree).find(key,trans);		
							} else if (rand < read + insert) {
								// System.out.println(Thread.currentThread().getName()+":
								// Insert "+key);
								trans = Runtime.createTransacation("insert "+key);
								trans.read(tree).insert(key, key, trans);
							} else {
								// System.out.println(Thread.currentThread().getName()+":
								// Delete "+key);
								trans = Runtime.createTransacation("delete "+key);
								trans.read(tree).delete(key, trans);
							}
							trans.commit();
							done = true;
						} catch(StmReadConflictException e) {
							bd.incRead();
							done = false;
						} catch (StmWriteConflictException e) {
							bd.incWrite();
							done = false;
						} catch (StmException e) {
							assert(false);
						}
					}
				}
		}
	}

	public static void main(String[] args) {
		
		NUM_THREADS = Integer.valueOf(args[0]);
		NUM_OPS = Integer.valueOf(args[1]);
		NUM_OPS_PER_THREAD = NUM_OPS / NUM_THREADS;

		/* updates are inserts and deletes */
		UPDATES = Integer.valueOf(args[2]) * 0.01;
		LOOKUPS = 1.0 - UPDATES;

		GAO<AVLTree<Integer, Integer>> tree = initTree();
		
		System.out.println("Starting " + NUM_THREADS + " worker threads.");
		SortedAdd threads[] = new SortedAdd[NUM_THREADS];

		for (int i = 0; i < NUM_THREADS; ++i) {
			BenchmarkData bd = new BenchmarkData();
			threads[i] = new SortedAdd(tree, LOOKUPS, UPDATES * 0.5,
					UPDATES * 0.5, bd);
			threads[i].setName("AppThread"+i);
		}

		long timer = System.currentTimeMillis();

		for (int i = 0; i < NUM_THREADS; ++i) {
			threads[i].start();
		}

		try {
			for (int i = 0; i < NUM_THREADS; ++i) {
				threads[i].join();
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

		timer = System.currentTimeMillis() - timer;

		int writeConflicts = 0;
		int readConflicts = 0;
		for (int i = 0; i < NUM_THREADS; ++i) {
			writeConflicts += threads[i].bd.getWriteConflicts();
			readConflicts += threads[i].bd.getReadConflicts();
		}

		try {
			FileWriter f = new FileWriter("stats.txt", true);
			f.append(Integer.toString(NUM_THREADS) + " ");
			f.append(Integer.toString(NUM_OPS) + " ");
			f.append(Double.toString(UPDATES) + " ");
			f.append(Long.toString(timer) + " ");
			f.append(Integer.toString(writeConflicts) + " ");
			f.append(Integer.toString(readConflicts) + " \n");
			f.flush();
			f.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		System.out.println("Write conflicts: " + writeConflicts);
		System.out.println("Read conflicts:  " + readConflicts);

	}

	private static GAO<AVLTree<Integer,Integer>> initTree() {
		GAO<AVLTree<Integer,Integer>> tree = null;
		boolean done = false;
		while(!done) {
			try {
				Transaction trans = Runtime.createTransacation("create");
				tree = GAO.create(new AVLTree<Integer,Integer>(),trans);
				trans.commit();
				done = true;
			} catch(StmException e) {
				done = false;
			}
		}
		// fill it with random numbers
		System.out.println("Filling Tree with " + NUM_INIT + " initial values.");
			done = false;
			while(!done) {
				try {
					Transaction trans = Runtime.createTransacation("init");
					for (int i = 0; i < NUM_INIT; ++i) {
						int key = (int) (Math.random() * RANGE);
						trans.read(tree).insert(key, key,trans);
					}
					trans.commit();
					done = true;
				} catch(StmException e) {
					done = false;
				}
			}
		return tree;
	}

}
