/*Copyright (C) 2009  Annette Bieniusa, Thomas Fuhrmann.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package dstm;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * GAO versions reflect the actual instances of the globally accessible objects.
 * 
 * @author Thomas Fuhrmann <fuhrmann@net.in.tum.de>
 * @author Annette Bieniusa <bieniusa@informatik.uni-freiburg.de>
 */
public class GAOV<T> {

	/**
	 * the corresponding GAO of which this is a version of
	 */
	private GAO<T> gao = null;
	private GAOV<T> predecessor = null;
	private Transaction writtenBy = null;
	
	private Set<Transaction> readingTransactions = new HashSet<Transaction>();
	private Set<GAOV<T>> successors = new HashSet<GAOV<T>>();
	
	/**
	 * The Double object(!) denotes the probability which the GAOV had during the last commit protocol round.
	 * If it is null, the GAOV has not participated yet in a protocol.
	 */
	private Double probability = null;
	/**
	 *  This dummy object serves as a lock for the successors set.
	 *  The successor set shouldn't be public.
	 */
	//public Object commitLock = new Object();
	
	protected GAO<T> getGao() {
		return gao;
	}

	protected GAOV<T> getPredecessor() {
		return predecessor;
	}

	protected Transaction getWrittenBy() {
		return writtenBy;
	}

	
	protected void setProbability(double probability) {
		this.probability = probability;
	}

	protected Double getProbability() {
		return probability;
	}

	/**
	 * Turn this instance into a GAO version. This is required if someone
	 * created a LOC that shall now become the initial GAO version.
	 * 
	 * @param gao
	 */
	protected void createFrom(GAO<T> gao, Transaction t) {
		this.gao = gao;
		writtenBy = t;
	}
	
	
	/**
	 * @return check set of this version
	 */
	protected Map<GAO<?>,GAOV<?>> getCheckSet() {
	  return writtenBy.getCheckSet();
	}

	/**
	 * The given transaction reads this GAO version. Return a local object
	 * copy that can be turned into a new GAO version upon commit of that
	 * transaction.
	 * 
	 * Only application threads shall call this method.
	 * 
	 * Called from
	 * - Transaction.read(GAO)
	 * 
	 * @return local object copy of this GAO version
	 */
	@SuppressWarnings("unchecked")
	protected GAOV<T> read(Transaction t) {

	  synchronized (readingTransactions) {	
		  readingTransactions.add(t);
	  }
	  try {

		  GAOV<T> loc = (GAOV<T>) super.clone();
		  
		  loc.successors = new HashSet<GAOV<T>>();
		  loc.predecessor = this;
		  loc.writtenBy = t;

		  loc.readingTransactions = new HashSet<Transaction>();
	
		  return loc;
		  
	  } catch (CloneNotSupportedException e) {
		  throw new RuntimeException("The VM says 'Clone not supported' even though we implement 'Cloneable'.");
	  }
	}

	/**
	 * Remove the given transaction from the 'readingTransactionSet' because the
	 * transaction is about to commit.
	 * 
	 * Only application threads shall call this method.
	 * 
	 * Called from - Transaction.commit()
	 * 
	 * No synchronization needed because even though multiple transactions may
	 * commit at the same time, we use the quick hack global commit lock in the
	 * calling method.
	 * 
	 */
	protected void removeFromReadingTransactionSet(Transaction t) {
		synchronized(readingTransactions) {
			readingTransactions.remove(t);
		}
	}

	
	/**
	 * Need to check whether there are delayed voting result messages in the set of GAO,
	 */
	
	protected GAOV<T> checkForSuccessfulPendingGAOV(Set<GAOV<?>> readSet) {	
		for (GAOV<?> readSetGAOV : readSet) {
			Map<GAO<?>,GAOV<?>> writeSet = readSetGAOV.writtenBy.getWriteSet();
			
			GAOV<?> wentry = writeSet.get(gao);
			if (wentry != null && this.isOlder(wentry)) {
				return (GAOV<T>) wentry;
			}
		}
		return null;
	}

	
	
	/**
	 * Check if the given read set contains a GAO version that dates before any
	 * of the GAO versions in the check set of this GAO version. If so, the
	 * calling transaction (i.e. the one that provided the given read set) would
	 * date before this GAO version and must not read this version.
	 * 
	 * @param readSet
	 * @return
	 * @throws StmReadConflictException 
	 */
	protected boolean inconsistentWith(Set<GAOV<?>> readSet) throws StmReadConflictException {

		Map<GAO<?>, GAOV<?>>	checkSet = writtenBy.getCheckSet();
		
		for ( GAOV<?> readSetGAOV : readSet ) {
			
			GAOV<?> checkSetGAOV = checkSet.get(readSetGAOV.gao);
			
			if ( checkSetGAOV != null ) {
				
					if ( readSetGAOV.isOlder(checkSetGAOV) ) return true;
				
			}
		}
		return false;
	}

	/**
	 * Return true if this GAO version is older than the given GAO version.
	 * 
	 * @param checkSetGAOV
	 * @return
	 */
	protected boolean isOlder(GAOV<?> checkSetGAOV) {
		while ( checkSetGAOV.predecessor != null ) {
			checkSetGAOV = checkSetGAOV.predecessor;
			if ( checkSetGAOV == this ) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Add a gaov to be written into the successor list.
	 * The caller needs to grab the commitLock object before 
	 * calling this function.
	 * 
	 * @param gaov GAOV to be inserted
	 */
	protected void addToSuccessors(GAOV<T> gaov) {
		successors.add(gaov);
	}
	
	/**
	 * Remove a gaov from the successor list.
	 * The caller needs to grab the commitLock object before 
	 * calling this function.
	 * 
	 * @param gaov GAOV to be removed
	 */
	
	protected void removeFromSuccessors(GAOV<T> gaov) {
		successors.remove(gaov);
	}
	
	/**
	 * Returns a copied set of all the GAOVs that lost during commit.
	 * @param gaov GAOV that is excluded
	 */
	
	protected Set<GAOV<T>> getAllSuccessorsBut(GAOV<T> gaov) {
		Set<GAOV<T>> copy = new HashSet<GAOV<T>> (successors);
		copy.remove(gaov);
		return copy;
	}
	
	protected boolean hasSuccessors() {
		return (!successors.isEmpty());
	}

	protected Set<GAOV<T>> getSuccessors() {
		return new HashSet<GAOV<T>> (successors);
	}

	
	
}
