/*Copyright (C) 2009  Annette Bieniusa, Thomas Fuhrmann.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package dstm;

import java.io.FileWriter;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * This class specifies at which thread a GAO registers.
 * The threads model the nodes in the final setting.
 * 
 * @author Annette Bieniusa
 */

public class Runtime implements Runnable {
	static final private int numberOfGaoThreads = 8;    //Maximum Number of GAO Threads
	static private Runtime[] runtimePool = new Runtime[numberOfGaoThreads];
	
	/**
	 * Queue of messages sent by transactions.
	 */
	private BlockingQueue<TransactionMessage> messageQueue = new LinkedBlockingQueue<TransactionMessage>();
	
	/**
	 * Hash table of all registered GAOs for this runtime.
	 */
	private Set<GAO<?>> registeredGaos = new HashSet<GAO<?>>();

	
	private NanoTimeCollector timer = new NanoTimeCollector();
	
	/**
	 * Manages also the mapping of transactions to threads. 
	 * This information is needed to get flat nesting of calls to atomic methods.
	 */
	private static ThreadLocal<Transaction> transaction = new ThreadLocal<Transaction>() {
		protected  Transaction initialValue() {
			return null;
		}
	};
	public static  Transaction getTransaction () {
		return transaction.get();
	}
	
	public static  Transaction createTransacation (String name) {
		Transaction t = new Transaction(name);
		transaction.set(t);
		return t;
	}
	
	public static void resetTransaction () {
		transaction.remove();
	}
	
	
	/**
	 * Runtimes get initialized lazily.
	 * @param g
	 */
	static public synchronized void registerGAO (GAO<?> g) {
		int rtId = g.hashCode() % numberOfGaoThreads;
		Runtime rt = runtimePool[rtId]; 
		
		if (rt == null) {
			rt = new Runtime(rtId);
			runtimePool[rtId] = rt;
		}
		rt.addGao(g);
	}
	
	/**
	 * Dispatches the message to the right runtime.
	 * @param m
	 */
	static synchronized protected void send(TransactionMessage m) {
		runtimePool[m.getReceiver().hashCode() % numberOfGaoThreads].enqueue(m);
	}
	
	
	private  void addGao(GAO<?> g) {
		synchronized (registeredGaos) {
			registeredGaos.add(g);
		}
	}
	
	private Runtime (int rtId) {
		Thread t = new Thread(this,"Runtime"+rtId);
		t.setDaemon(true);
		t.start();
	}
	
	public void run() {
		while (true) {
			try {
				TransactionMessage m = (TransactionMessage) messageQueue.take();
				//System.out.println(Thread.currentThread().getId() +" : "+this + "received :" + m);
				timer.start();
				GAO<?> receiver = m.getReceiver();
				receiver.consume(m);
				timer.stop();
				try {
					FileWriter f = new FileWriter("bench/" + Thread.currentThread().getName() +".txt", true);
					String s = "";
					s += System.currentTimeMillis() + " ";
					s += timer.getTotalDuration()+ " ";
					s += m + " ";
					f.append(s +"\n");
					f.flush();
					f.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			timer.reset();
		}
		
	}
	
	/**
	 * Each GAO has an associated message queue. The runtime thread checks 
	 * for new incoming messages and processes them. Since we are using 
	 * a blocking queue, the access to the queue is synchronized.
	 * @param m Message that will be enqueued
	 */
	protected void enqueue(TransactionMessage m) {
		try {
			messageQueue.put(m);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
