/*Copyright (C) 2009  Annette Bieniusa, Thomas Fuhrmann.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package dstm;

import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * The Transaction class reflects the state that is associate with a
 * transaction. Transaction instances survive the actual transaction even some
 * time after a successful commit of the transaction, because they contain
 * information about the relation among the GAO versions.
 * 
 * @author Thomas Fuhrmann <fuhrmann@net.in.tum.de>
 * @author Annette Bieniusa <bieniusa@informatik.uni-freiburg.de>
 * 
 */
public class Transaction {

	private NanoTimeCollector appTimer = new NanoTimeCollector();
	private NanoTimeCollector commitTimer = new NanoTimeCollector();
	private long id;
	private String description;
	private int roundsInCommitProtocol = 0;

	private Set<GAOV<? extends Object>> readSet = new HashSet<GAOV<? extends Object>>();
	private Map<GAO<? extends Object>, GAOV<? extends Object>> localGaovHeap = new HashMap<GAO<? extends Object>, GAOV<? extends Object>>();
	private Map<GAO<? extends Object>, GAOV<? extends Object>> writeSet = new HashMap<GAO<? extends Object>, GAOV<? extends Object>>();
	private Map<GAO<? extends Object>, GAOV<? extends Object>> checkSet = new HashMap<GAO<? extends Object>, GAOV<? extends Object>>();
	private Map<GAO<? extends Object>, GAOV<? extends Object>> createSet = new HashMap<GAO<? extends Object>, GAOV<? extends Object>>();

	private BlockingQueue<GaoMessage<?>> messageQueue = new LinkedBlockingQueue<GaoMessage<?>>();

	private static enum State {
		OPEN, SUCCESS, FAILED
	};

	State state = State.OPEN;

	/*
	 * The default constructor for transaction is only accessible within the
	 * dstm package. Hence, all transactions are created via the
	 * Runtime.createTransaction() method. This gives inter-operability with
	 * code that was transactified via Bytecode rewriting.
	 */
	protected Transaction(String description) {
		// FIXME Is this clever?
		this.id = this.hashCode();
		this.description = description;
		appTimer.start();
	}

	/**
	 * Return a new local object copy for the given GAO that that is cloned from
	 * a GAO version that does not conflict with the read set of this
	 * transaction.
	 * 
	 * Only application threads shall call this method.
	 * 
	 * @throws StmException
	 */

	@SuppressWarnings("unchecked")
	/*
	 * The cast of GAOV<T> to T is valid by construction.
	 */
	public <T> T read(GAO<T> gao) {

		if (gao == null) {
			return null;
		}

		/*
		 * Check locally if the GAO is already in the read set of this
		 * transaction. If so, return the respective version.
		 */
		GAOV<?> readGaov = localGaovHeap.get(gao);
		if (readGaov != null) {
			return (T) readGaov;
		}

		/*
		 * Check locally if the GAO has just been created.
		 */
		readGaov = createSet.get(gao);
		if (readGaov != null) {
			return (T) readGaov;
		}

		/*
		 * Ask the GAO now to send a clone of an appropriate non-conflicting
		 * version.
		 */
		appTimer.stop();
		Runtime.send(new ReadMessage(this, gao, readSet));

		try {
			GaoMessage m = messageQueue.take();
			// System.out.println(this + "received :" + m);
			switch (m.getType()) {
			case SENDGAOV:
				SendGAOVMessage<T> sm = (SendGAOVMessage<T>) m;
				GAOV<T> gaov = sm.gaov;
				GAOV<T> loc = sm.loc;

				readSet.add(gaov);
				localGaovHeap.put(gao, loc);

				mergeIntoCheckSet(gaov.getCheckSet());
				appTimer.start();
				return (T) loc;

			default:
				new RuntimeException("Transaction " + this.toString()
						+ " received the wrong kind of message!");
			}

		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		return null;
	}

	private void mergeIntoCheckSet(Map<GAO<?>, GAOV<?>> otherSet) {
		for (Entry<GAO<?>, GAOV<?>> entry : otherSet.entrySet()) {
			GAOV<?> mostRecent = checkSet.get(entry.getKey());
			if (mostRecent == null) {
				checkSet.put(entry.getKey(), entry.getValue());
			} else {
				if (mostRecent.isOlder(entry.getValue())) {
					checkSet.put(entry.getKey(), entry.getValue());
				}
			}
		}

	}

	/**
	 * Add the given LOC to the write set of this transaction.
	 * 
	 * Only application threads shall call this method.
	 * 
	 * @param gaov
	 */
	public <T> void write(GAOV<T> gaov) {
		if (!createSet.containsValue(gaov) && !writeSet.containsValue(gaov)) {
			writeSet.put(gaov.getGao(), gaov);
		}
	}

	/**
	 * The transaction has finished executing. Now the commit phase is started.
	 * Only application threads shall call this method.
	 * 
	 * @throws StmException
	 */

	public void commit() throws StmException {
		appTimer.stop();
		commitTimer.start();
		/*
		 * A new round of the commit protocol gets executed.
		 */
		state = State.OPEN;

		unregisterFromReadSet();
		/*
		 * Notify all the GAOs which this transaction has written that a new
		 * version should be entered into the pending version tree. The GAOs
		 * then start or continue their commit protocol.
		 */
		for (Entry<GAO<?>, GAOV<?>> entry : writeSet.entrySet()) {
			Runtime.send(new CommitMessage(this, entry.getKey(), entry
					.getValue()));
		}

		while (state == State.OPEN) {
			++roundsInCommitProtocol;
			/*
			 * The transaction now collects the votes of all GAOS which are
			 * involved in this commit.
			 */

			VotingResult result = collectVotes();

			/*
			 * Depending on the outcome, i.e. the collected votes, notify the
			 * GAOs. result = -1.0 encodes failure, result = 2.0 encodes
			 * success. Other results encode an open outcome.
			 */

			sendVotingResult(result);
		}
		commitTimer.stop();
		// Output for benchmarks

		try {
			FileWriter f = new FileWriter("bench/"
					+ Thread.currentThread().getName() + ".txt", true);
			String s = "";
			s += System.currentTimeMillis() + " ";
			s += id + " ";
			s += state + " ";
			s += appTimer.getTotalDuration() + " ";
			s += commitTimer.getTotalDuration() + " ";
			s += roundsInCommitProtocol + " ";
			s += readSet.size() + " ";
			s += writeSet.size() + " ";
			s += createSet.size() + " ";
			s += description + " ";
			f.append(s + "\n");
			f.flush();
			f.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		// System.out.println(this+" is in state "+state);
		switch (state) {
		case SUCCESS:
			// System.out.println(this + " has read set: "+localGaovHeap);
			// System.out.println(this + " has write set: "+writeSet);
			break;
		case FAILED:
			throw new StmWriteConflictException();
		}

	}

	/**
	 * Remove this transaction from all the read GAO versions' 'reading
	 * transaction set'.
	 */
	protected void unregisterFromReadSet() {
		for (GAOV<?> gaov : readSet) {
			gaov.removeFromReadingTransactionSet(this);
		}
	}

	private void sendVotingResult(VotingResult result) {
		for (Entry<GAO<?>, GAOV<?>> entry : writeSet.entrySet()) {
			Runtime.send(new VotingResultMessage(this, entry.getKey(), entry
					.getValue(), result));
		}
	}

	/**
	 * The given GAO version has received a positive or negative vote. We now
	 * count the votes for all GAO versions in our write set. As long a at least
	 * one has not received a vote yet, the decision is still open. If all GAO
	 * versions have received a positive vote the decision is made and we have
	 * been selected.
	 * 
	 * Only the commit protocol's worker thread shall call this method.
	 * 
	 */

	private VotingResult collectVotes() {
		int selected = 0;
		int openVotes = writeSet.size();
		boolean failed = false;

		while (openVotes > 0) {
			try {
				commitTimer.stop();
				VoteMessage<?> m = (VoteMessage<?>) messageQueue.take();
				commitTimer.start();
				// System.out.println(this + "received :" + m);
				switch (m.getVote()) {
				case POS:
					selected++;
					openVotes--;
					break;
				case NEG:
					openVotes--;
					break;
				case FAILED:
					failed = true;
					openVotes--;
					break;
				default:
					new RuntimeException("Transaction " + this.toString()
							+ " received an invalid vote.");
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		/*
		 * CAREFUL: Don't change the order of these two statements! Read-only
		 * transactions have selected == 0 but they should always commit
		 * successfully.
		 */

		// TODO Can we make a "short-cut" for read-only transactions?
		if (selected == writeSet.size()) {
			state = State.SUCCESS;

			checkSet.putAll(writeSet);

			// create all the GAOs that where created during this transaction
			for (GAO<?> gao : createSet.keySet()) {
				Runtime.registerGAO(gao);
			}
			return VotingResult.SUCCESS;
		} else if (failed || selected == 0) {
			state = State.FAILED;
			return VotingResult.FAILED;
		}

		return new VotingResult(((double) selected) / (double) writeSet.size());
	}

	protected Map<GAO<?>, GAOV<?>> getCheckSet() {
		return checkSet;

	}

	protected Map<GAO<?>, GAOV<?>> getWriteSet() {
		return writeSet;
	}

	/**
	 * Add the GAO to the set of GAOs to be accessible globally if the
	 * transaction commits successfully.
	 * 
	 * @param gao
	 * @return
	 */
	protected void addToCreateSet(GAO<?> gao, GAOV<?> loc) {
		createSet.put(gao, loc);
	}

	/**
	 * Enqueue a message in the transaction's message queue. This function is
	 * thread safe as we use a blocking queue.
	 */
	protected void enqueue(GaoMessage<?> m) {
		try {
			messageQueue.put(m);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
